/*
 * Copyright (c) 2017, Red Hat Inc. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package org.openjdk.gcbench.fragger;

import org.openjdk.jmh.annotations.*;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.TimeUnit;


@Warmup(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Measurement(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Fork(value = 1)
@BenchmarkMode(Mode.AverageTime)
@OutputTimeUnit(TimeUnit.NANOSECONDS)
@Threads(1)
@State(Scope.Benchmark)
public class CHMFragger {

    static final int PAYLOAD_SIZE = 4096;

    @Param({"1000"})
    int size;

    Object[] keys;
    Map<Object, Object> cache;

    @Setup
    public void setup() {
        keys = new Object[size];
        cache = new ConcurrentHashMap<>();
        for (int c = 0; c < size; c++) {
            Object k = new Object();
            keys[c] = k;
            cache.put(k, new byte[PAYLOAD_SIZE]);
        }
    }

    @Benchmark
    public Object test() {
        ThreadLocalRandom tlr = ThreadLocalRandom.current();
        int idx = tlr.nextInt(size);
        if (tlr.nextInt(5) == 0) {
            Object pk = keys[idx];
            Object nk = new Object();
            byte[] v = new byte[PAYLOAD_SIZE];
            cache.put(nk, v);
            cache.remove(pk);
            keys[idx] = nk;
            return v;
        } else {
            return cache.get(keys[idx]);
        }
    }


}
